from datetime import datetime
from functools import lru_cache

import magic
from django.conf import settings
from django.http.response import HttpResponseBase
from django.utils.http import http_date

from .models import FileModel
from .utils import get_all_subclasses

DJANGO_STORAGE_BACKEND = "django.core.files.storage.FileSystemStorage"


class BlobResponseBuilder:
    storage_backend: str

    @classmethod
    def get_content_type(cls, file_record: FileModel) -> str:
        return NotImplemented

    @classmethod
    def get_last_modified(cls, file_record: FileModel) -> int:
        return NotImplemented

    @classmethod
    def get_file_name(cls, file_record: FileModel) -> str:
        return NotImplemented

    @classmethod
    def get_content_length(cls, file_record: FileModel) -> int:
        return NotImplemented

    @classmethod
    @lru_cache
    def get_response_builder(cls):
        current_storage = getattr(settings, "STORAGES", {}).get("default", {}).get("BACKEND") or DJANGO_STORAGE_BACKEND

        return next(
            (subclass for subclass in get_all_subclasses(cls) if subclass.storage_backend == current_storage),
            DefaultStorageResponseBuilder,
        )

    @classmethod
    def build_response[T: HttpResponseBase](cls, file_record: FileModel, base_response: T) -> T:
        response_builder = cls.get_response_builder()

        base_response["Content-Type"] = response_builder.get_content_type(file_record)
        base_response["Last-Modified"] = http_date(response_builder.get_last_modified(file_record))
        base_response["Content-Disposition"] = f'inline; filename="{response_builder.get_file_name(file_record)}"'
        base_response["Content-Length"] = response_builder.get_content_length(file_record)

        return base_response


class DefaultStorageResponseBuilder(BlobResponseBuilder):
    storage_backend = DJANGO_STORAGE_BACKEND

    @classmethod
    def get_content_type(cls, file_record: FileModel):
        return magic.from_buffer(file_record.file.file.read(2048))

    @classmethod
    def get_last_modified(cls, file_record: FileModel):
        return datetime.now().timestamp()

    @classmethod
    def get_file_name(cls, file_record: FileModel):
        return file_record.file.file.name

    @classmethod
    def get_content_length(cls, file_record: FileModel):
        return file_record.file.size


class GoogleStorageResponseBuilder(DefaultStorageResponseBuilder):
    storage_backend = "storages.backends.gcloud.GoogleCloudStorage"

    @classmethod
    def get_content_type(cls, file_record: FileModel):
        return file_record.file.file.mime_type

    @classmethod
    def get_last_modified(cls, file_record: FileModel):
        return file_record.uploaded_at.timestamp()


class S3Boto3StorageResponseBuilder(GoogleStorageResponseBuilder):
    storage_backend = "storages.backends.s3boto3.S3Boto3Storage"

    @classmethod
    def get_content_type(cls, file_record: FileModel):
        return magic.from_buffer(file_record.file.file.read(2048))



# Generated by Django 5.1.4 on 2024-12-08 22:57

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flex_blob', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='filemodel',
            name='uploaded_at',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now, verbose_name='Uploaded at'),
            preserve_default=False,
        ),
    ]



import datetime

from django.http import (
    HttpRequest,
    HttpResponseBadRequest,
    HttpResponseForbidden,
    StreamingHttpResponse,
)
from django.utils.translation import gettext_lazy as _
from django.views import View

from .builders import BlobResponseBuilder
from .models import FileModel


class MediaView(View):
    def get(self, request: HttpRequest, path: str):
        if not path:
            return HttpResponseBadRequest(_("Invalid file path."))

        if file_record := FileModel.objects.filter(file=path).first():
            if not (request.user and request.user.is_staff) and not file_record.check_auth(request):
                return HttpResponseBadRequest(_("Invalid file path."))
        else:
            file_record = FileModel(file=path, uploaded_at=datetime.datetime.now())

        return self.serve_file(file_record)

    def serve_file(self, file_record: FileModel, chunk_size: int = 8192):
        def file_iterator():
            with file_record.file.open("rb") as file_obj:
                while chunk := file_obj.read(chunk_size):
                    yield chunk

        try:
            response = StreamingHttpResponse(file_iterator())
            return BlobResponseBuilder.build_response(file_record, response)
        except FileNotFoundError:
            return HttpResponseBadRequest(_("Invalid file path."))



from django.db import models


class DynamicFileField(models.FileField):
    def clean(self, value, model_instance):
        cleaned_value = super().clean(value, model_instance)

        if validators := getattr(model_instance, "get_validators", None):
            assert callable(validators)
            for validator in validators():
                validator(cleaned_value)

        return cleaned_value



def get_all_subclasses(klass):
    subclasses = set()
    work = [klass]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                work.append(child)
    return subclasses



import os
from typing import List

from django.core.validators import BaseValidator
from django.db import models
from django.utils.translation import gettext_lazy as _

from .fields import DynamicFileField


def dynamic_upload_to(instance, filename):
    return instance.get_upload_path(filename)


class FileModel(models.Model):
    file = DynamicFileField(
        upload_to=dynamic_upload_to,
        verbose_name=_("File"),
        db_index=True,
    )
    uploaded_at = models.DateTimeField(auto_now_add=True, verbose_name=_("Uploaded at"))

    def check_auth(self, request=None):
        subclass_instance = self.get_subclass_instance()
        if subclass_instance and hasattr(subclass_instance, "check_auth"):
            return subclass_instance.check_auth(request)
        return False

    def get_upload_path(self, filename):
        return os.path.join("uploads", filename)

    def get_subclass_instance(self):
        for related_object in self._meta.related_objects:
            if hasattr(self, related_object.name):
                return getattr(self, related_object.name, None)
        return None

    def get_validators[T: BaseValidator](self) -> List[T]:
        return []

    def __str__(self):
        return self.file.name

    class Meta:
        verbose_name = _("Base File")
        verbose_name_plural = _("Base Files")



# Generated by Django 5.1.3 on 2024-12-03 09:17

import flex_blob.fields
import flex_blob.models
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='FileModel',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('file', flex_blob.fields.DynamicFileField(db_index=True, upload_to=flex_blob.models.dynamic_upload_to, verbose_name='File')),
            ],
            options={
                'verbose_name': 'Base File',
                'verbose_name_plural': 'Base Files',
            },
        ),
    ]
